% Based on Pan cardiac action potential BG modelling 2018
% only LCC channel is retained.

clear;
clc;
close all;

%% Set directories
current_dir = cd;
Idx_backslash = find(current_dir == filesep);
% main_dir = current_dir(1:Idx_backslash(end));
data_dir = [current_dir filesep 'data' filesep];
code_dir = [current_dir filesep 'code' filesep];
output_dir = [current_dir filesep 'output' filesep];
storage_dir = [current_dir filesep 'storage' filesep];

%% Define constants
R = 8.314; % unit J/mol/K
T = 310;
F = 96485;

% cNao_st = 140; % unit mM
% cNai_st = 10; % unit mM
% cKo_st = 5.4;
% cKi_st = 145;
% 
N_A = 6.022e23;
% 
A_cap = 1.534e-4; % Unit cm^2

x_LCC = 50000/N_A*1e15; % unit fmol
% x_Na_channel = 122720/N_A*1e15; % unit fmol
% x_K1_channel = 4261/N_A*1e15; % unit fmol
% x_K_channel = 5369/N_A*1e15; % unit fmol
% x_Kp_channel = 725/N_A*1e15; % unit fmol

%% Define volumes (unit pL)
W_i = 38;
W_e = 5.182;

% %% Import IV parameters
% % Na channel
% G_Na = 16*A_cap; % Unit mA/V
% E_Na = R*T/F*log(cNao_st/cNai_st);
% E_Na_norm = F*E_Na/(R*T);
% G_GHK = 2*G_Na*(1-exp(E_Na_norm))/(cNai_st - cNao_st*exp(E_Na_norm))*R*T/F; % Unit mA/mM
% P_Na = G_GHK/F * 1e12; % Unit pL/s

% L-type calcium channel
% Load P_GHK (unit pL/s)
load([storage_dir 'LCC_P_GHK_Ca.mat']);
P_LCC_Ca = P_GHK; % Unit pL/s
load([storage_dir 'LCC_P_GHK_K.mat']);
P_LCC_K = P_GHK; % Unit pL/s

% % K1 channel
% load([storage_dir 'K1_G_GHK.mat']);
% P_K1 = G_GHK/F * 1e12; % Unit pL/s
% 
% % K channel
% % Load G_GHK
% load([storage_dir 'K_G_GHK.mat']);
% P_K = G_GHK/F * 1e12; % Unit pL/s
% 
% % Kp channel
% load([storage_dir 'Kp_G_GHK.mat']);
% P_Kp = G_GHK/F * 1e12; % Unit pL/s

% %% Import gate transition parameters
% load([storage_dir 'Na_m_parameters.mat']);
% params_m = params_vec;
% 
% load([storage_dir 'Na_h_parameters.mat']);
% params_h = params_vec;
% 
% load([storage_dir 'Na_j_parameters.mat']);
% params_j = params_vec;
% 
% load([storage_dir 'K1_parameters.mat']);
% params_K1 = params_vec;
% 
% load([storage_dir 'K_X_parameters.mat']);
% params_X = params_vec;
% 
% % Calculate parameters for X_i gate
% A = exp(-56.26/32.1);
% K = 1/(1+A*exp(-120/32.1));

% alpha0 = K; % Unit ms^-1
% beta0 = A*K; % Unit ms^-1
% 
% zf = 0;
% zr = R*T/32.1/F;
% params_Xi = [alpha0; zf*1e3; beta0; zr*1e3];

% A = exp(7.488/5.98);
% K = 1/(1+A*exp(-60/5.98));
% 
% alpha0 = K; % Unit ms^-1
% beta0 = A*K; % Unit ms^-1
% 
% zf = 0;
% zr = -R*T/5.98/F;
% 
% params_Kp = [alpha0; zf*1e3; beta0; zr*1e3];
% 
load([storage_dir 'LCC_d_parameters.mat']);
params_d = params_vec;

load([storage_dir 'LCC_f_parameters.mat']);
params_f = params_vec;

% %% Load Na/K pump kinetic parameters
% load([data_dir 'NaK_kinetic_fitting_results.mat']);
% params = params_vec;
% 
% fast_kinetic_constant = 1e6;
% 
% K_d_Nai = params(3);
% k_3b_p = fast_kinetic_constant;
% k_3b_m = fast_kinetic_constant*(0.5*K_d_Nai);
% 
% k_4b_p = fast_kinetic_constant;
% k_4b_m = fast_kinetic_constant*(2*K_d_Nai);
% 
% K_d_Nai_0 = params(1);
% k_5b_m_0 = fast_kinetic_constant;
% k_5b_p_0 = fast_kinetic_constant/K_d_Nai_0;
% 
% k_6b_p = params(7);
% k_6b_m = k_6b_p/6.3;
% 
% k_7b_p = params(8);
% k_7b_m = params(11);
% 
% K_d_Nae_0 = params(2);
% k_8b_m_0 = fast_kinetic_constant;
% k_8b_p_0 = fast_kinetic_constant*K_d_Nae_0;
% 
% K_d_Nae = params(4);
% k_9b_m = fast_kinetic_constant;
% k_9b_p = fast_kinetic_constant*2*K_d_Nae;
% 
% k_10b_m = fast_kinetic_constant;
% k_10b_p = fast_kinetic_constant*0.5*K_d_Nae;
% 
% K_d_Ke = params(5);
% k_11b_p = fast_kinetic_constant;
% k_11b_m = fast_kinetic_constant*0.5*K_d_Ke;
% 
% k_12b_p = fast_kinetic_constant;
% k_12b_m = fast_kinetic_constant*2*K_d_Ke;
% 
% k_13b_p = params(9);
% k_13b_m = params(12);
% 
% K_d_MgATP = params(6);
% k_14b_p = fast_kinetic_constant;
% k_14b_m = fast_kinetic_constant*K_d_MgATP;
% 
% k_15b_p = params(10);
% k_15b_m = params(13);

% % Calculate remaining parameter
% G_MgATP_0 = 11900;
% RT = 8.314*310;
% K_MgATP = exp(-G_MgATP_0/RT)*10^6;
% K_d_Ki = sqrt((k_3b_m*k_4b_m*k_5b_m_0*k_6b_m*k_7b_m*k_8b_m_0*k_9b_m*k_10b_m*k_11b_m*k_12b_m*k_13b_m*k_14b_m*k_15b_m*K_MgATP)/(k_3b_p*k_4b_p*k_5b_p_0*k_6b_p*k_7b_p*k_8b_p_0*k_9b_p*k_10b_p*k_11b_p*k_12b_p*k_13b_p*k_14b_p*k_15b_p));
% 
% k_1b_m = fast_kinetic_constant;
% k_1b_p = fast_kinetic_constant*2*K_d_Ki;
% 
% k_2b_m = fast_kinetic_constant;
% k_2b_p = fast_kinetic_constant*0.5*K_d_Ki;
% 
% Delta_NaK = params(14);

% %% Calculate NCX kinetic parameters (in terms of concentration)
% load([storage_dir 'NCX_fitting_results.mat']);
% kappa_fast = 1e6;
% kappa_3 = params_vec(1);
% kappa_6 = params_vec(2);
% K_1 = params_vec(3);
% K_2 = params_vec(4);
% K_3 = params_vec(5);
% K_4 = params_vec(6);
% K_5 = params_vec(7);
% K_6 = params_vec(8);
% 
% K_Nai = W_i*params_vec(9);
% K_Cai = W_i*params_vec(10);
% K_Nae = K_Nai;
% K_Cae = K_Cai;
% 
% Delta_NCX = params_vec(11);
% 
% kf_NCX = [kappa_fast*K_1;...
%     kappa_fast*K_2*K_Cai;...
%     kappa_3*K_3;...
%     kappa_fast*K_4;...
%     kappa_fast*K_5*K_Nae^3;...
%     kappa_6*K_6];
% kr_NCX = [kappa_fast*K_2*K_Nai^3;...
%     kappa_fast*K_3;...
%     kappa_3*K_4;...
%     kappa_fast*K_5*K_Cae;...
%     kappa_fast*K_6;...
%     kappa_6*K_1];

% %% Kinetics of calcium buffering
% fast_kinetic_constant = 1e6;
% 
% Km_TRPN = 0.5e-3; % unit mM
% Km_CMDN = 2.38e-3; % unit mM
% 
% kf_TRPN = fast_kinetic_constant;
% kr_TRPN = fast_kinetic_constant*Km_TRPN;
% 
% kf_CMDN = fast_kinetic_constant;
% kr_CMDN = fast_kinetic_constant*Km_CMDN;
% 
% kf_Ca_buffer = [kf_TRPN; kf_CMDN];
% kr_Ca_buffer = [kr_TRPN; kr_CMDN];

%% Load stoichiometric matrices
stoich_path = [data_dir 'LCC_forward_matrix.txt'];
stoich_file_id = fopen(stoich_path,'r');

stoich_file_data = textscan(stoich_file_id,'%s','delimiter','\n');
fclose(stoich_file_id);

num_rows = length(stoich_file_data{1});
num_cols = sum(stoich_file_data{1}{1} == ',')+1;

N_f = zeros(num_rows,num_cols);

for i_row = 1:num_rows
    line_str = stoich_file_data{1}{i_row};
    line_split = regexp(line_str,',','split');
    
    for i_col = 1:num_cols
        N_f(i_row,i_col) = str2double(line_split{i_col});
    end
end

% Load reverse matrix
stoich_path = [data_dir 'LCC_reverse_matrix.txt'];
stoich_file_id = fopen(stoich_path,'r');

stoich_file_data = textscan(stoich_file_id,'%s','delimiter','\n');
fclose(stoich_file_id);

num_rows = length(stoich_file_data{1}); % num of kappa + K
num_cols = sum(stoich_file_data{1}{1} == ',')+1; % num of reactions

N_r = zeros(num_rows,num_cols);

for i_row = 1:num_rows
    line_str = stoich_file_data{1}{i_row};
    line_split = regexp(line_str,',','split');
    
    for i_col = 1:num_cols
        N_r(i_row,i_col) = str2double(line_split{i_col});
    end
end

N_fT = transpose(N_f);
N_rT = transpose(N_r);

N = N_r - N_f;
N_T = N_rT - N_fT;

num_cols = size(N,2);
I = eye(num_cols);

M = [I N_fT; I N_rT];
M_rref = rref(M);

M_T = transpose(M);
M_T_rref = rref(M_T);

%% Set up the vectors

% % K1 channel
% alpha_K1_bg = params_K1(1)*1e3; % unit s^-1
% beta_K1_bg = params_K1(3)*1e3; % unit s^-1
% 
% kf_K1 = [P_K1/x_K1_channel/sqrt(cKo_st); ... % R_GHK
%     alpha_K1_bg]; % Rg
% 
% kr_K1 = [P_K1/x_K1_channel/sqrt(cKo_st); ... % R_GHK
%     beta_K1_bg]; % Rg
% 
% % K channel
% alpha_X_bg = params_X(1); % unit s^-1
% beta_X_bg = params_X(3); % unit s^-1
% 
% alpha_Xi_bg = params_Xi(1)*1e3; % unit s^-1
% beta_Xi_bg = params_Xi(3)*1e3; % unit s^-1
% 
% kf_K = [P_K/x_K_channel/sqrt(cKo_st); ... % R_GHK
%     2*alpha_X_bg; ... % RX00
%     alpha_X_bg; ... % RX00
%     2*alpha_X_bg; ... % RX00
%     alpha_X_bg; ... % RX00
%     alpha_Xi_bg; ... % RX00
%     alpha_Xi_bg; ... % RX00
%     alpha_Xi_bg]; % RX00
% 
% kr_K = [P_K/x_K_channel/sqrt(cKo_st); ... % R_GHK
%     beta_X_bg; ... % RX00
%     2*beta_X_bg; ... % RX00
%     beta_X_bg; ... % RX00
%     2*beta_X_bg; ... % RX00
%     beta_Xi_bg; ... % RX00
%     beta_Xi_bg; ... % RX00
%     beta_Xi_bg]; % RX00
% 
% % Kp channel
% alpha_Kp_bg = params_Kp(1)*1e3; % unit s^-1
% beta_Kp_bg = params_Kp(3)*1e3; % unit s^-1
% 
% kf_Kp = [P_Kp/x_Kp_channel; ... % R_GHK
%     alpha_Kp_bg]; % Rg
% 
% kr_Kp = [P_Kp/x_Kp_channel; ... % R_GHK
%     beta_Kp_bg]; % Rg
% 
% % Na channel
% alpha_m0_bg = params_m(1)*1e3; % unit s^-1
% beta_m0_bg = params_m(3)*1e3; % unit s^-1
% 
% alpha_h0_bg = params_h(1)*1e3; % unit s^-1
% beta_h0_bg = params_h(3)*1e3; % unit s^-1
% 
% alpha_j0_bg = params_j(1)*1e3; % unit s^-1
% beta_j0_bg = params_j(3)*1e3; % unit s^-1
% 
% kf_Na = [P_Na/x_Na_channel; ... % R_GHK
%     3*alpha_m0_bg; ... % Rm000
%     3*alpha_m0_bg; ... % Rm001
%     3*alpha_m0_bg; ... % Rm010
%     3*alpha_m0_bg; ... % Rm011
%     2*alpha_m0_bg; ... % Rm100
%     2*alpha_m0_bg; ... % Rm101
%     2*alpha_m0_bg; ... % Rm110
%     2*alpha_m0_bg; ... % Rm111
%     alpha_m0_bg; ... % Rm200
%     alpha_m0_bg; ... % Rm201
%     alpha_m0_bg; ... % Rm210
%     alpha_m0_bg; ... % Rm211
%     alpha_h0_bg; ... % Rh000
%     alpha_h0_bg; ... % Rh001
%     alpha_h0_bg; ... % Rh100
%     alpha_h0_bg; ... % Rh101
%     alpha_h0_bg; ... % Rh200
%     alpha_h0_bg; ... % Rh201
%     alpha_h0_bg; ... % Rh300
%     alpha_h0_bg; ... % Rh301
%     alpha_j0_bg; ... % Rj000
%     alpha_j0_bg; ... % Rj010
%     alpha_j0_bg; ... % Rj100
%     alpha_j0_bg; ... % Rj110
%     alpha_j0_bg; ... % Rj200
%     alpha_j0_bg; ... % Rj210
%     alpha_j0_bg; ... % Rj300
%     alpha_j0_bg]; % Rj310
% 
% kr_Na = [P_Na/x_Na_channel; ... % R_GHK
%     beta_m0_bg; ... % Rm000
%     beta_m0_bg; ... % Rm001
%     beta_m0_bg; ... % Rm010
%     beta_m0_bg; ... % Rm011
%     2*beta_m0_bg; ... % Rm100
%     2*beta_m0_bg; ... % Rm101
%     2*beta_m0_bg; ... % Rm110
%     2*beta_m0_bg; ... % Rm111
%     3*beta_m0_bg; ... % Rm200
%     3*beta_m0_bg; ... % Rm201
%     3*beta_m0_bg; ... % Rm210
%     3*beta_m0_bg; ... % Rm211
%     beta_h0_bg; ... % Rh000
%     beta_h0_bg; ... % Rh001
%     beta_h0_bg; ... % Rh100
%     beta_h0_bg; ... % Rh101
%     beta_h0_bg; ... % Rh200
%     beta_h0_bg; ... % Rh201
%     beta_h0_bg; ... % Rh300
%     beta_h0_bg; ... % Rh301
%     beta_j0_bg; ... % Rj000
%     beta_j0_bg; ... % Rj010
%     beta_j0_bg; ... % Rj100
%     beta_j0_bg; ... % Rj110
%     beta_j0_bg; ... % Rj200
%     beta_j0_bg; ... % Rj210
%     beta_j0_bg; ... % Rj300
%     beta_j0_bg]; % Rj310

% L-type calcium channel
alpha_d0_bg = params_d(1)*1e3; % unit s^-1
beta_d0_bg = params_d(3)*1e3; % unit s^-1

alpha_f1_0_bg = params_f(1)*1e3; % unit s^-1
beta_f1_0_bg = params_f(3)*1e3; % unit s^-1

alpha_f2_0_bg = params_f(5)*1e3; % unit s^-1
beta_f2_0_bg = params_f(7)*1e3; % unit s^-1

K_f3_0 = beta_f1_0_bg*alpha_f2_0_bg/alpha_f1_0_bg/beta_f2_0_bg; % dimensionless
rate_f3 = 1e5; % unit s^-1

KmCa = 0.6e-3; % Unit mM
rate_fCa = 1e5; % unit s^-1

kf_LCC = [P_LCC_Ca/x_LCC; ... % R_GHK_Ca1
    P_LCC_Ca/x_LCC; ... % R_GHK_Ca2
    P_LCC_K/x_LCC; ... % R_GHK_K1
    P_LCC_K/x_LCC; ... % R_GHK_K2
    alpha_d0_bg; ... % Rd000
    alpha_d0_bg; ... % Rd010
    alpha_d0_bg; ... % Rd020
    alpha_d0_bg; ... % Rd001
    alpha_d0_bg; ... % Rd011
    alpha_d0_bg; ... % Rd021
    alpha_f1_0_bg; ... % Rf1_000
    alpha_f1_0_bg; ... % Rf1_100
    alpha_f1_0_bg; ... % Rf1_001
    alpha_f1_0_bg; ... % Rf1_101
    alpha_f2_0_bg; ... % Rf2_000
    alpha_f2_0_bg; ... % Rf2_100
    alpha_f2_0_bg; ... % Rf2_001
    alpha_f2_0_bg; ... % Rf2_101
    K_f3_0*rate_f3; ... % Rf3_010
    K_f3_0*rate_f3; ... % Rf3_110
    K_f3_0*rate_f3; ... % Rf3_011
    K_f3_0*rate_f3; ... % Rf3_111
    rate_fCa; ... % RfCa000
    rate_fCa; ... % RfCa100
    rate_fCa; ... % RfCa010
    rate_fCa; ... % RfCa110
    rate_fCa; ... % RfCa020
    rate_fCa]; % RfCa120

kr_LCC = [P_LCC_Ca/x_LCC; ... % R_GHK_Ca1
    P_LCC_Ca/x_LCC; ... % R_GHK_Ca2
    P_LCC_K/x_LCC; ... % R_GHK_K1
    P_LCC_K/x_LCC; ... % R_GHK_K2
    beta_d0_bg; ... % Rd000
    beta_d0_bg; ... % Rd010
    beta_d0_bg; ... % Rd020
    beta_d0_bg; ... % Rd001
    beta_d0_bg; ... % Rd011
    beta_d0_bg; ... % Rd021
    beta_f1_0_bg; ... % Rf1_000
    beta_f1_0_bg; ... % Rf1_100
    beta_f1_0_bg; ... % Rf1_001
    beta_f1_0_bg; ... % Rf1_101
    beta_f2_0_bg; ... % Rf2_000
    beta_f2_0_bg; ... % Rf2_100
    beta_f2_0_bg; ... % Rf2_001
    beta_f2_0_bg; ... % Rf2_101
    rate_f3; ... % Rf3_010
    rate_f3; ... % Rf3_110
    rate_f3; ... % Rf3_011
    rate_f3; ... % Rf3_111
    rate_fCa/KmCa; ... % RfCa000
    rate_fCa/KmCa; ... % RfCa100
    rate_fCa/KmCa; ... % RfCa010
    rate_fCa/KmCa; ... % RfCa110
    rate_fCa/KmCa; ... % RfCa020
    rate_fCa/KmCa]; % RfCa120

% % Na/K pump
% kf_NaK = transpose([k_1b_p k_2b_p k_3b_p k_4b_p k_5b_p_0 k_6b_p k_7b_p k_8b_p_0 ...
%     k_9b_p k_10b_p k_11b_p k_12b_p k_13b_p k_14b_p k_15b_p]);
% kr_NaK = transpose([k_1b_m k_2b_m k_3b_m k_4b_m k_5b_m_0 k_6b_m k_7b_m k_8b_m_0 ...
%     k_9b_m k_10b_m k_11b_m k_12b_m k_13b_m k_14b_m k_15b_m]);

% Overall
k = [kf_LCC;kr_LCC];
W = [ones(size(N,2),1);W_i;W_e;W_i;W_e;ones(12,1)]; % kappa Cai Cae Ki Ke [gates]

lambdaW = exp(pinv(M)*log(k));
lambda = lambdaW./W;
kappa = lambda(1:size(N,2));
K = lambda(size(N,2)+1:end);

save([output_dir 'LCC_parameters.mat'],'kappa','K','params_f');

%% Checks
N_rref = rref(N);
R_mat = null(N,'r');

K_eq = kf_LCC./kr_LCC;
zero_est = R_mat'*K_eq;

k_est = exp(M*log(lambdaW));
diff = sum(abs((k-k_est)./k));

%% output for cellML
if true
    kappa = lambda(1:num_cols);
    K = lambda(num_cols+1:end);

    fID = fopen([data_dir 'rxnID.txt'], 'r');
    rxnID = textscan(fID,'%s', 'delimiter','\n');
    fclose(fID);
    fID = fopen([data_dir 'Kname.txt'], 'r');
    Kname = textscan(fID,'%s');
    fclose(fID);

    for j = 1:length(kappa)
        fprintf('var kappa_%s: fmol_per_sec {init: %g};\n',rxnID{1}{j},kappa(j));
    end
    for j = 1:length(K)
        fprintf('var K_%s: per_fmol {init: %g};\n',Kname{1}{j},K(j));
    end

    if false
        qname = {'L','LR','LRG','LR_BARK','LRG_BARK','R','RG','B1d','B1p','PKACI_B1p',...
'B1tot','PKACI','BARK','G','GaT','GBy','GaD','GsGDP','Pi'};
        qval = [0.0001672850,0.0000016970,0.0001046605,0.0000016970,0.0001046605,0.0002197247,...
            0.0000254602,0.0000000000,...
0.0000219260,.0000219260,0.0004579000,0.0022120882,0.0228000000,0.1453052187,...
0.0009519000,0.0009762200,0.0000244948,0.0000000000,570]';
        for j=1:length(qname)
            fprintf('var q_%s: fmol {init: %g};\n',qname{j}, qval(j));
        end
    end

    for j=1:length(K)
        fprintf('var q_%s: fmol {init: 1e-13};\n',Kname{1}{j});
    end

    for j=1:length(kappa)
        fprintf('var v%s: fmol_per_sec;\n',rxnID{1}{j})
    end
    for j=1:length(K)
        fprintf('var mu_%s: J_per_mol;\n',Kname{1}{j});
    end
    for j=1:length(K)
        fprintf('mu_%s = R*T*ln(K_%s*q_%s);\n',Kname{1}{j},Kname{1}{j},Kname{1}{j});
    end
    for j=1:length(kappa)
        fprintf('v%s = kappa_%s*exp(mu_a/(R*T));\n',rxnID{1}{j}, rxnID{1}{j})
    end
    for j=1:length(K)
        fprintf('ode(q_%s, time) = p;\n',Kname{1}{j});
    end

end
